//
// includes
//

#include <map>
#include <list>
#include <iostream>


#include "pad.h"
#include "synaptics.h"
#include "synparams.h"
#include "syndebug.h"

#include "version.h"

#include <X11/Xdefs.h> // for typedef Bool
#include <sys/shm.h>
#include <stdlib.h>


#define PAROFF( verId, par ) \
    offsetof( SynShm, verId ) + offsetof( ShmSegment##verId, par )

#define SETPAR( verId, p, par, t, min, max ) \
    p.offset = PAROFF( verId, par ); \
    p.type = t; \
    p.min_val = min; \
    p.max_val = max;

#define NODRIVERVER VER( 0, 0, 0 )


//
// local varibles
//

bool Synaptics::Pad::mHasShm = false;
Synaptics::DriverVersion Synaptics::Pad::mDetectedDriver = DV_UNKNOWN;


std::map<std::string, Synaptics::Pad::Parameter> Synaptics::Pad::mSupportedDriver;
std::list<std::string> Synaptics::Pad::mSupportedDriverList;
::SynShm* Synaptics::Pad::mSynShm = NULL;

Synaptics::Pad *Synaptics::Pad::mSelf = NULL;


Synaptics::Pad::Pad()
{
    SYNDEBUG( "Pad" );
    
    // NOTE better don't call anything except the register routine!

    //
    // register supported driver parameters
    //
    registerParameters();
}


Synaptics::Pad::~Pad()
{
    SYNDEBUG( "~Pad" );

    if ( mSynShm )
    {
        SYNDEBUG( "detach driver" );
        shmdt( mSynShm );
    }
}

void Synaptics::Pad::INIT()
{
    SYNDEBUG( "INIT" );

    if ( !mSelf )
        mSelf = new Synaptics::Pad(); // call constructor
    else
        return;

    //
    // NOTE important! Otherwise the destructor is not being called!
    //
    atexit( Synaptics::Pad::CLOSE );

    SYNDEBUG( "This version of libsynaptics supports the following driver versions:" );

    std::list<std::string>::iterator it = mSupportedDriverList.begin();

    for ( it =  mSupportedDriverList.begin(); it !=  mSupportedDriverList.end(); it++ )
        SYNDEBUG( " --- "<< *it );

    //
    // detect driver version
    //
    detectDriverVersion();

    //
    // attach proper segment
    //
    mHasShm = false;
    int sizeOfShm;

    switch( driverVersion() )
    {
        case ( NODRIVERVER ):
            SYNERR( "cannot access driver, please install and configure it properly!" );
            mDetectedDriver = DV_UNKNOWN;
            return;

        case ( VER( 0, 14, 4 ) ):
            sizeOfShm = sizeof( ShmSegment_0_14_4 );
            break;

        //
        // TODO insert upcoming driver versions
        //

        default:
            SYNERR( "outdated or unknown driver version, please update at least to version 0.14.4!" );
            mDetectedDriver = DV_OUTDATED;
            return;
    }

    mDetectedDriver = DV_SUPPORTED;

    // reconnect to the shared memory area
    int shmid;

    if ( ( shmid = shmget( SHM_SYNAPTICS, sizeOfShm, 0 ) ) == -1 )
    {
        if ( ( shmid = shmget( SHM_SYNAPTICS, 0, 0 ) ) == -1 )
        {
            SYNERR( "can't access shared memory area. SHMConfig disabled?" );
        }
        else
        {
            SYNERR( "incorrect size of shared memory area. Incompatible driver version?" );
        }
    }
    else
    {
        // attach
        if ( ( mSynShm = ( SynShm* ) shmat( shmid, NULL, 0 ) ) == NULL)
        {
            perror( "shmat" );
        }
        {
            SYNDEBUG( "class is operateable" );
            mHasShm = true;
        }
    }

}

void Synaptics::Pad::CLOSE()
{
    SYNDEBUG( "CLOSE" );
    
    if ( mSelf ) delete mSelf; // call destructor
    mSelf = 0;
}


bool Synaptics::Pad::detectDriverVersion()
{
    SYNDEBUG( "detectDriverVersion" );
    
    int shmid;

    mSynShm = NULL;

    // connect to the shared memory area
    if ( ( shmid = shmget( SHM_SYNAPTICS, sizeof( ShmSegment_Version ), 0 ) ) == -1 )
    {
        if ( ( shmid = shmget( SHM_SYNAPTICS, 0, 0 ) ) == -1 )
        {
            SYNERR( "can't access shared memory area. SHMConfig disabled?" )
        }
        else
        {
            SYNERR( "incorrect size of shared memory area. Incompatible driver version?" );
        }
    }
    else
    {
        // attach
        if ( ( mSynShm = ( ::SynShm* ) shmat( shmid, NULL, SHM_RDONLY ) ) == NULL)
        {
            perror( "shmat" );
        }
        else
        {
            SYNDEBUG( "detected driver version " << driverStr() );
            return true;
        }

    }

    return false;
}

bool Synaptics::Pad::hasHardware()
{
    SYNDEBUG( "hasHardware" );
    INIT();
    SYNERR( "IMPLEMENT ME" );
    return false;
}

Synaptics::DriverVersion Synaptics::Pad::driverKind()
{
    SYNDEBUG( "driverKind" );
    INIT();
    return mDetectedDriver;
}

bool Synaptics::Pad::hasDriver()
{
    SYNDEBUG( "hasDriver" );
    INIT();
    return ( DV_SUPPORTED == driverKind() );
}

bool Synaptics::Pad::hasSynClient()
{
    SYNDEBUG( "hasSynClient" );
    SYNERR( "IMPLEMENT ME" );
    return false;
}

bool Synaptics::Pad::hasSynDaemon()
{
    SYNDEBUG( "hasSynDaemon" );
    SYNERR( "IMPLEMENT ME" );
    return false;
}

bool Synaptics::Pad::isSynDaemonRunning()
{
    SYNDEBUG( "isSynDaemonRunning" );
    SYNERR( "IMPLEMENT ME" );
    return false;
}

bool Synaptics::Pad::killSynDaemonProcesses()
{
    SYNDEBUG( "killSynDaemonProcesses" );
    SYNERR( "IMPLEMENT ME" );
    return false;
}

std::string Synaptics::Pad::intVerToStrVer( int v )
{
    SYNDEBUG( "intVerToStrVer" );
    char ver[10];

    if ( v == 0)
        return std::string( "None" );

    sprintf(
        ver,
        "%d.%d.%d",
        v / 10000,
        ( v / 100 ) % 100,
        v % 100
    );

    return std::string( ver );

}

std::string Synaptics::Pad::driverStr()
{
    SYNDEBUG( "driverStr" );
    INIT();
    return intVerToStrVer( driverVersion() );
}

std::string Synaptics::Pad::libraryStr()
{
    SYNDEBUG( "libraryStr" );
    INIT();
    return LIBSYNAPTICS_VERSIONSTR;
}

Synaptics::PadType Synaptics::Pad::getPadType()
{
    SYNDEBUG( "getPadType" );
    INIT();

    PadType pt = PT_NONE;

    // FIXME: detection just based on the assumption that the driver will register as unknown
    if ( hasShm() and driverKind() == DV_UNKNOWN )
        pt = PT_UNKNOWN;
    else
    {
        // TODO: decide between ALPS, SYNAPTICS and MULTISYN
        pt = PT_MULTISYN;
    }
    
    return pt;
}

bool Synaptics::Pad::hasParam( std::string param )
{
    SYNDEBUG( "hasParam" );
    INIT();

    return
        mSupportedDriver[ driverStr() ].find( param ) !=
        mSupportedDriver[ driverStr() ].end();
}

double Synaptics::Pad::getParam( std::string param )
{
    SYNDEBUG( "getParam" );
    INIT();

    Bool rb = 0;
    int ri = 0;
    double rd = 0;


    if ( hasParam( param ) )
    {
        Param p = mSupportedDriver[ driverStr() ][ param ];

        if ( ( p.type == PT_BOOL ) or ( p.type == PT_BOOL_RO ) )
        {
            rb = *(Bool*)(((char*)mSynShm) + p.offset );
            return rb;
        }

        if ( ( p.type == PT_INT ) or ( p.type == PT_INT_RO ) )
        {
            ri = *(int*)(((char*)mSynShm) + p.offset );
            return ri;
        }

        if ( ( p.type == PT_DOUBLE ) or ( p.type == PT_DOUBLE_RO ) )
        {
            rd = *(double*)(((char*)mSynShm) + p.offset );
            return rd;
        }
    }
    else
         return 0;
}

void Synaptics::Pad::setParam( std::string param, double v )
{
    SYNDEBUG( "setParam" );
    INIT();

    if ( hasParam( param ) )
    {
        Param p = mSupportedDriver[ driverStr() ][ param ];

        if ( ( p.min_val <= v ) and ( p.max_val >= v ) )
        {
            if ( p.type == PT_BOOL )
                (*(Bool*)(((char*)mSynShm) + p.offset)) = (Bool)v;

            if ( p.type == PT_INT )
                (*(int*)(((char*)mSynShm) + p.offset)) = (int)v;

            if ( p.type == PT_DOUBLE )
                (*(double*)(((char*)mSynShm) + p.offset)) = (double)v;

            mSupportedDriver[ driverStr() ][ param ] = p;
        }
    }
}

int Synaptics::Pad::driverVersion()
{
    SYNDEBUG( "driverVersion" );
    INIT();

    return mSynShm? mSynShm->ver.version : VER( 0, 0, 0);
}

bool Synaptics::Pad::hasShm()
{
    SYNDEBUG( "hasShm" );
    INIT();

    return mHasShm;
}

double Synaptics::Pad::getParamMin( std::string param )
{
    SYNDEBUG( "getParamMin" );
    INIT();

    if ( hasParam( param ) )
    {
        Param p = mSupportedDriver[ driverStr() ][ param ];

        return p.min_val;
    }
    else
    {
        return 0;
    }
}

double Synaptics::Pad::getParamMax( std::string param )
{
    SYNDEBUG( "getParamMax" );
    INIT();

    if ( hasParam( param ) )
    {
        Param p = mSupportedDriver[ driverStr() ][ param ];

        return p.max_val;
    }
    else
    {
        return 0;
    }
}

Synaptics::ParamType Synaptics::Pad::getParamType( std::string param )
{
    SYNDEBUG( "getParamType" );
    INIT();

    if ( hasParam( param ) )
    {
        Param p = mSupportedDriver[ driverStr() ][ param ];

        return p.type;
    }
    else
    {
        return Synaptics::PT_VOID;
    }
}

std::list<std::string> Synaptics::Pad::getAvailableParams()
{
    SYNDEBUG( "getAvailableParams" );
    INIT();

    std::list<std::string> myParams;

    Parameter p = mSupportedDriver[ driverStr() ];

    for ( Parameter::iterator it = p.begin(); it != p.end(); it++ )
    {
        myParams.insert( myParams.end(), it->first );
    }


    return myParams;
}


//
// shared shm for all synaptics driver versions
// TODO add new driver version here!
// NOTE you MUST NOT call INIT() from within that method!
//

void Synaptics::Pad::registerParameters()
{
    SYNDEBUG( "registerParameters" );
    Parameter p;
    //std::cout << "registerParameters" << std::endl;
    //
    // register supported versions
    //

    SETPAR( _0_14_4, p[ LEFTEDGE ],                 left_edge,                  PT_INT,     0,     10000 );
    SETPAR( _0_14_4, p[ RIGHTEDGE ],                right_edge,                 PT_INT,     0,     10000 );
    SETPAR( _0_14_4, p[ TOPEDGE ],                  top_edge,                   PT_INT,     0,     10000 );
    SETPAR( _0_14_4, p[ BOTTOMEDGE ],               bottom_edge,                PT_INT,     0,     10000 );
    SETPAR( _0_14_4, p[ FINGERLOW ],                finger_low,                 PT_INT,     0,     255 );
    SETPAR( _0_14_4, p[ FINGERHIGH ],               finger_high,                PT_INT,     0,     255 );
    SETPAR( _0_14_4, p[ MAXTAPTIME ],               tap_time,                   PT_INT,     0,     1000 );
    SETPAR( _0_14_4, p[ MAXTAPMOVE ],               tap_move,                   PT_INT,     0,     2000 );
    SETPAR( _0_14_4, p[ MAXDOUBLETAPTIME ],         tap_time_2,                 PT_INT,     0,     1000 );
    SETPAR( _0_14_4, p[ CLICKTIME ],                click_time,                 PT_INT,     0,     1000 );
    SETPAR( _0_14_4, p[ FASTTAPS ],                 fast_taps,                  PT_BOOL,    0,     1 );
    SETPAR( _0_14_4, p[ EMULATEMIDBUTTONTIME ],     emulate_mid_button_time,    PT_INT,     0,     1000 );
    SETPAR( _0_14_4, p[ VERTSCROLLDELTA ],          scroll_dist_vert,           PT_INT,     0,     1000 );
    SETPAR( _0_14_4, p[ HORIZSCROLLDELTA ],         scroll_dist_horiz,          PT_INT,     0,     1000 );
    SETPAR( _0_14_4, p[ MINSPEED ],                 min_speed,                  PT_DOUBLE,  0,     1.0 );
    SETPAR( _0_14_4, p[ MAXSPEED ],                 max_speed,                  PT_DOUBLE,  0,     1.0 );
    SETPAR( _0_14_4, p[ ACCELFACTOR ],              accl,                       PT_DOUBLE,  0,     0.2 );
    SETPAR( _0_14_4, p[ EDGEMOTIONMINZ ],           edge_motion_min_z,          PT_INT,     1,     255 );
    SETPAR( _0_14_4, p[ EDGEMOTIONMAXZ ],           edge_motion_max_z,          PT_INT,     1,     255 );
    SETPAR( _0_14_4, p[ EDGEMOTIONMINSPEED ],       edge_motion_min_speed,      PT_INT,     0,     1000 );
    SETPAR( _0_14_4, p[ EDGEMOTIONMAXSPEED ],       edge_motion_max_speed,      PT_INT,     0,     1000 );
    SETPAR( _0_14_4, p[ EDGEMOTIONUSEALWAYS ],      edge_motion_use_always,     PT_BOOL,    0,     1 );
    SETPAR( _0_14_4, p[ UPDOWNSCROLLING ],          updown_button_scrolling,    PT_BOOL,    0,     1 );
    SETPAR( _0_14_4, p[ LEFTRIGHTSCROLLING ],       leftright_button_scrolling, PT_BOOL,    0,     1 );
    SETPAR( _0_14_4, p[ UPDOWNREPEAT ],             updown_button_repeat,       PT_BOOL,    0,     1 );
    SETPAR( _0_14_4, p[ LEFTRIGHTREPEAT ],          leftright_button_repeat,    PT_BOOL,    0,     1 );
    SETPAR( _0_14_4, p[ SCROLLBUTTONREPEAT ],       scroll_button_repeat,       PT_INT,     SBR_MIN, SBR_MAX );
    SETPAR( _0_14_4, p[ TOUCHPADOFF ],              touchpad_off,               PT_INT,     0,     2 );
    SETPAR( _0_14_4, p[ GUESTMOUSEOFF ],            guestmouse_off,             PT_BOOL,    0,     1 );
    SETPAR( _0_14_4, p[ LOCKEDDRAGS ],              locked_drags,               PT_BOOL,    0,     1 );
    SETPAR( _0_14_4, p[ RTCORNERBUTTON ],           tap_action[RT_TAP],         PT_INT,     0,     SYN_MAX_BUTTONS );
    SETPAR( _0_14_4, p[ RBCORNERBUTTON ],           tap_action[RB_TAP],         PT_INT,     0,     SYN_MAX_BUTTONS );
    SETPAR( _0_14_4, p[ LTCORNERBUTTON ],           tap_action[LT_TAP],         PT_INT,     0,     SYN_MAX_BUTTONS );
    SETPAR( _0_14_4, p[ LBCORNERBUTTON ],           tap_action[LB_TAP],         PT_INT,     0,     SYN_MAX_BUTTONS );
    SETPAR( _0_14_4, p[ TAPBUTTON1 ],               tap_action[F1_TAP],         PT_INT,     0,     SYN_MAX_BUTTONS );
    SETPAR( _0_14_4, p[ TAPBUTTON2 ],               tap_action[F2_TAP],         PT_INT,     0,     SYN_MAX_BUTTONS );
    SETPAR( _0_14_4, p[ TAPBUTTON3 ],               tap_action[F3_TAP],         PT_INT,     0,     SYN_MAX_BUTTONS );
    SETPAR( _0_14_4, p[ CIRCULARSCROLLING ],        circular_scrolling,         PT_BOOL,    0,     1 );
    SETPAR( _0_14_4, p[ CIRCSCROLLDELTA ],          scroll_dist_circ,           PT_DOUBLE,  0.01,  3 );
    SETPAR( _0_14_4, p[ CIRCSCROLLTRIGGER ],        circular_trigger,           PT_INT,     0,     8 );
    SETPAR( _0_14_4, p[ CIRCULARPAD ],              circular_pad,               PT_BOOL,    0,     1 );
    SETPAR( _0_14_4, p[ PALMDETECT ],               palm_detect,                PT_BOOL,    0,     1 );
    SETPAR( _0_14_4, p[ PALMMINWIDTH ],             palm_min_width,             PT_INT,     0,     15 );
    SETPAR( _0_14_4, p[ PALMMINZ ],                 palm_min_z,                 PT_INT,     0,     255 );
    SETPAR( _0_14_4, p[ COASTINGSPEED ],            coasting_speed,             PT_DOUBLE,  0,     20 );
    SETPAR( _0_14_4, p[ ABSCOORDX ],                x,                          PT_INT_RO,  1,     0 ); // read only parameter!
    SETPAR( _0_14_4, p[ ABSCOORDY ],                y,                          PT_INT_RO,  1,     0 ); // read only parameter!
    SETPAR( _0_14_4, p[ VERSIONNUMBER ],            version,                    PT_INT_RO,  1,     0 ); // read only parameter!

    mSupportedDriver[ "0.14.4" ] = p;
    mSupportedDriverList.insert( mSupportedDriverList.end(), "0.14.4" );

    // TODO make an entry for upcoming versions :)
};

